/*
 * This file was generated automatically by ExtUtils::ParseXS version 3.57 from the
 * contents of Fast.xs. Do not edit this file, edit Fast.xs instead.
 *
 *    ANY CHANGES MADE HERE WILL BE LOST!
 *
 */

#line 1 "lib/GIS/Distance/Fast.xs"
#define PERL_NO_GET_CONTEXT

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include "ppport.h"

#include <math.h>

const double DEG_RADS = M_PI / 180.;
const double KILOMETER_RHO = 6371.64;

/* Sphere with equal meridian length */
const double RM = 6367449.14582342;

/* WGS 84 Ellipsoid */
const double A = 6378137.;
const double B = 6356752.314245;
const double F = 1. / 298.257223563;

double alt_distance(double lat1, double lon1, double lat2, double lon2) {
    double f = 0.5 * (lat2 + lat1) * DEG_RADS;
    double g = 0.5 * (lat2 - lat1) * DEG_RADS;
    double l = 0.5 * (lon2 - lon1) * DEG_RADS;

    double sf = sin(f), sg = sin(g), sl = sin(l);
    double s2f = sf * sf, s2g = sg * sg, s2l = sl * sl;
    double c2f = 1. - s2f, c2g = 1. - s2g, c2l = 1. - s2l;

    double s2 = s2g * c2l + c2f * s2l;
    double c2 = c2g * c2l + s2f * s2l;

    double s, c, omega, rr, aa, bb, pp, qq, d2, qp, eps1, eps2;

    if (s2 == 0.) return 0.;
    if (c2 == 0.) return M_PI * RM / 0.001;

    s = sqrt(s2), c = sqrt(c2);
    omega = atan2(s, c);
    rr = s * c;
    aa = s2g * c2f / s2 + s2f * c2g / c2;
    bb = s2g * c2f / s2 - s2f * c2g / c2;
    pp = rr / omega;
    qq = omega / rr;
    d2 = s2 - c2;
    qp = qq + 6. * pp;
    eps1 = 0.5 * F * (-aa - 3. * bb * pp);
    eps2 = 0.25 * F * F * ((-qp * bb + (-3.75 + d2 * (qq + 3.75 * pp)) *
            aa + 4. - d2 * qq) * aa - (7.5 * d2 * bb * pp - qp) * bb);

    double d = 2. * omega * A * (1. + eps1 + eps2);
    return d * 0.001;
}

double cosine_distance(double lat1, double lon1, double lat2, double lon2) {
    lon1 *= DEG_RADS;
    lat1 *= DEG_RADS;
    lon2 *= DEG_RADS;
    lat2 *= DEG_RADS;

    double a = sin( lat1 ) * sin( lat2 );
    double b = cos( lat1 ) * cos( lat2 ) * cos( lon2 - lon1 );
    double c = acos( a + b );

    return KILOMETER_RHO * c;
}

double great_circle_distance(double lat1, double lon1, double lat2, double lon2) {
    lon1 *= DEG_RADS;
    lat1 *= DEG_RADS;
    lon2 *= DEG_RADS;
    lat2 *= DEG_RADS;

    double c = 2 * asin( sqrt(
        pow( sin((lat1-lat2)/2), 2.0 ) +
        cos(lat1) * cos(lat2) *
        pow( sin((lon1-lon2)/2), 2.0 )
    ) );

    return KILOMETER_RHO * c;
}

double haversine_distance(double lat1, double lon1, double lat2, double lon2) {
    lon1 *= DEG_RADS;
    lat1 *= DEG_RADS;
    lon2 *= DEG_RADS;
    lat2 *= DEG_RADS;

    double dlon = lon2 - lon1;
    double dlat = lat2 - lat1;
    double a = pow( sin(dlat/2.0), 2.0 ) + cos(lat1) * cos(lat2) * pow( sin(dlon/2.0), 2.0 );
    double c = 2.0 * atan2( sqrt(a), sqrt(fabs(1.0-a)) );

    return KILOMETER_RHO * c;
}

double null_distance(double lat1, double lon1, double lat2, double lon2) {
    return 0;
}

double polar_distance(double lat1, double lon1, double lat2, double lon2) {
    lon1 *= DEG_RADS;
    lat1 *= DEG_RADS;
    lon2 *= DEG_RADS;
    lat2 *= DEG_RADS;

    double a = M_PI / 2 - lat1;
    double b = M_PI / 2 - lat2;
    double c = sqrt( pow(a, 2.0) + pow(b, 2.0) - 2 * a * b * cos(lon2 - lon1) );

    return KILOMETER_RHO * c;
}

double vincenty_distance(double lat1, double lon1, double lat2, double lon2) {
    if ( (lon1==lon2) && (lat1==lat2) ) {
        return 0;
    }

    lon1 *= DEG_RADS;
    lat1 *= DEG_RADS;
    lon2 *= DEG_RADS;
    lat2 *= DEG_RADS;

    double a = 6378137.0;
    double b = 6356752.3142;
    double f = 1/298.257223563;

    double l = lon2 - lon1;

    double u1 = atan( (1-f) * tan(lat1) );
    double u2 = atan( (1-f) * tan(lat2) );

    double sin_u1 = sin(u1);
    double cos_u1 = cos(u1);
    double sin_u2 = sin(u2);
    double cos_u2 = cos(u2);

    double lambda     = l;
    double lambda_pi  = 2 * M_PI;
    int iter_limit = 20;

    double cos_sq_alpha = 0.0;
    double sin_sigma    = 0.0;
    double cos2sigma_m  = 0.0;
    double cos_sigma    = 0.0;
    double sigma        = 0.0;

    while( fabs(lambda-lambda_pi) > 1e-12 && --iter_limit>0 ){
        double sin_lambda = sin(lambda);
        double cos_lambda = cos(lambda);

        sin_sigma = sqrt((cos_u2*sin_lambda) * (cos_u2*sin_lambda) +
            (cos_u1*sin_u2-sin_u1*cos_u2*cos_lambda) * (cos_u1*sin_u2-sin_u1*cos_u2*cos_lambda));

        cos_sigma = sin_u1*sin_u2 + cos_u1*cos_u2*cos_lambda;
        sigma = atan2(sin_sigma, cos_sigma);

        double alpha = asin(cos_u1 * cos_u2 * sin_lambda / sin_sigma);
        cos_sq_alpha = cos(alpha) * cos(alpha);
        cos2sigma_m = cos_sigma - 2.0*sin_u1*sin_u2/cos_sq_alpha;

        double cc = f/16.0*cos_sq_alpha*(4.0+f*(4.0-3.0*cos_sq_alpha));
        lambda_pi = lambda;
        lambda = l + (1.0-cc) * f * sin(alpha) *
            (sigma + cc*sin_sigma*(cos2sigma_m+cc*cos_sigma*(-1.0+2.0*cos2sigma_m*cos2sigma_m)));
    }

    double usq = cos_sq_alpha*(a*a-b*b)/(b*b);
    double aa = 1.0 + usq/16384.0*(4096.0+usq*(-768.0+usq*(320.0-175.0*usq)));
    double bb = usq/1024.0 * (256.0+usq*(-128.0+usq*(74.0-47.0*usq)));
    double delta_sigma = bb*sin_sigma*(cos2sigma_m+bb/4.0*(cos_sigma*(-1.0+2.0*cos2sigma_m*cos2sigma_m)-
        bb/6.0*cos2sigma_m*(-3.0+4.0*sin_sigma*sin_sigma)*(-3.0+4.0*cos2sigma_m*cos2sigma_m)));
    double c = b*aa*(sigma-delta_sigma);

    return c / 1000;
}

#line 188 "lib/GIS/Distance/Fast.c"
#ifndef PERL_UNUSED_VAR
#  define PERL_UNUSED_VAR(var) if (0) var = var
#endif

#ifndef dVAR
#  define dVAR		dNOOP
#endif


/* This stuff is not part of the API! You have been warned. */
#ifndef PERL_VERSION_DECIMAL
#  define PERL_VERSION_DECIMAL(r,v,s) (r*1000000 + v*1000 + s)
#endif
#ifndef PERL_DECIMAL_VERSION
#  define PERL_DECIMAL_VERSION \
	  PERL_VERSION_DECIMAL(PERL_REVISION,PERL_VERSION,PERL_SUBVERSION)
#endif
#ifndef PERL_VERSION_GE
#  define PERL_VERSION_GE(r,v,s) \
	  (PERL_DECIMAL_VERSION >= PERL_VERSION_DECIMAL(r,v,s))
#endif
#ifndef PERL_VERSION_LE
#  define PERL_VERSION_LE(r,v,s) \
	  (PERL_DECIMAL_VERSION <= PERL_VERSION_DECIMAL(r,v,s))
#endif

/* XS_INTERNAL is the explicit static-linkage variant of the default
 * XS macro.
 *
 * XS_EXTERNAL is the same as XS_INTERNAL except it does not include
 * "STATIC", ie. it exports XSUB symbols. You probably don't want that
 * for anything but the BOOT XSUB.
 *
 * See XSUB.h in core!
 */


/* TODO: This might be compatible further back than 5.10.0. */
#if PERL_VERSION_GE(5, 10, 0) && PERL_VERSION_LE(5, 15, 1)
#  undef XS_EXTERNAL
#  undef XS_INTERNAL
#  if defined(__CYGWIN__) && defined(USE_DYNAMIC_LOADING)
#    define XS_EXTERNAL(name) __declspec(dllexport) XSPROTO(name)
#    define XS_INTERNAL(name) STATIC XSPROTO(name)
#  endif
#  if defined(__SYMBIAN32__)
#    define XS_EXTERNAL(name) EXPORT_C XSPROTO(name)
#    define XS_INTERNAL(name) EXPORT_C STATIC XSPROTO(name)
#  endif
#  ifndef XS_EXTERNAL
#    if defined(HASATTRIBUTE_UNUSED) && !defined(__cplusplus)
#      define XS_EXTERNAL(name) void name(pTHX_ CV* cv __attribute__unused__)
#      define XS_INTERNAL(name) STATIC void name(pTHX_ CV* cv __attribute__unused__)
#    else
#      ifdef __cplusplus
#        define XS_EXTERNAL(name) extern "C" XSPROTO(name)
#        define XS_INTERNAL(name) static XSPROTO(name)
#      else
#        define XS_EXTERNAL(name) XSPROTO(name)
#        define XS_INTERNAL(name) STATIC XSPROTO(name)
#      endif
#    endif
#  endif
#endif

/* perl >= 5.10.0 && perl <= 5.15.1 */


/* The XS_EXTERNAL macro is used for functions that must not be static
 * like the boot XSUB of a module. If perl didn't have an XS_EXTERNAL
 * macro defined, the best we can do is assume XS is the same.
 * Dito for XS_INTERNAL.
 */
#ifndef XS_EXTERNAL
#  define XS_EXTERNAL(name) XS(name)
#endif
#ifndef XS_INTERNAL
#  define XS_INTERNAL(name) XS(name)
#endif

/* Now, finally, after all this mess, we want an ExtUtils::ParseXS
 * internal macro that we're free to redefine for varying linkage due
 * to the EXPORT_XSUB_SYMBOLS XS keyword. This is internal, use
 * XS_EXTERNAL(name) or XS_INTERNAL(name) in your code if you need to!
 */

#undef XS_EUPXS
#if defined(PERL_EUPXS_ALWAYS_EXPORT)
#  define XS_EUPXS(name) XS_EXTERNAL(name)
#else
   /* default to internal */
#  define XS_EUPXS(name) XS_INTERNAL(name)
#endif

#ifndef PERL_ARGS_ASSERT_CROAK_XS_USAGE
#define PERL_ARGS_ASSERT_CROAK_XS_USAGE assert(cv); assert(params)

/* prototype to pass -Wmissing-prototypes */
STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params);

STATIC void
S_croak_xs_usage(const CV *const cv, const char *const params)
{
    const GV *const gv = CvGV(cv);

    PERL_ARGS_ASSERT_CROAK_XS_USAGE;

    if (gv) {
        const char *const gvname = GvNAME(gv);
        const HV *const stash = GvSTASH(gv);
        const char *const hvname = stash ? HvNAME(stash) : NULL;

        if (hvname)
	    Perl_croak_nocontext("Usage: %s::%s(%s)", hvname, gvname, params);
        else
	    Perl_croak_nocontext("Usage: %s(%s)", gvname, params);
    } else {
        /* Pants. I don't think that it should be possible to get here. */
	Perl_croak_nocontext("Usage: CODE(0x%" UVxf ")(%s)", PTR2UV(cv), params);
    }
}
#undef  PERL_ARGS_ASSERT_CROAK_XS_USAGE

#define croak_xs_usage        S_croak_xs_usage

#endif

/* NOTE: the prototype of newXSproto() is different in versions of perls,
 * so we define a portable version of newXSproto()
 */
#ifdef newXS_flags
#define newXSproto_portable(name, c_impl, file, proto) newXS_flags(name, c_impl, file, proto, 0)
#else
#define newXSproto_portable(name, c_impl, file, proto) (PL_Sv=(SV*)newXS(name, c_impl, file), sv_setpv(PL_Sv, proto), (CV*)PL_Sv)
#endif /* !defined(newXS_flags) */

#if PERL_VERSION_LE(5, 21, 5)
#  define newXS_deffile(a,b) Perl_newXS(aTHX_ a,b,file)
#else
#  define newXS_deffile(a,b) Perl_newXS_deffile(aTHX_ a,b)
#endif

/* simple backcompat versions of the TARGx() macros with no optimisation */
#ifndef TARGi
#  define TARGi(iv, do_taint) sv_setiv_mg(TARG, iv)
#  define TARGu(uv, do_taint) sv_setuv_mg(TARG, uv)
#  define TARGn(nv, do_taint) sv_setnv_mg(TARG, nv)
#endif

#line 339 "lib/GIS/Distance/Fast.c"

XS_EUPXS(XS_GIS__Distance__Fast_alt_distance); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_GIS__Distance__Fast_alt_distance)
{
    dVAR; dXSARGS;
    if (items != 4)
       croak_xs_usage(cv,  "lat1, lon1, lat2, lon2");
    {
	double	lat1 = (double)SvNV(ST(0))
;
	double	lon1 = (double)SvNV(ST(1))
;
	double	lat2 = (double)SvNV(ST(2))
;
	double	lon2 = (double)SvNV(ST(3))
;
	double	RETVAL;
	dXSTARG;

	RETVAL = alt_distance(lat1, lon1, lat2, lon2);
	TARGn((double)RETVAL, 1);
	ST(0) = TARG;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_GIS__Distance__Fast_cosine_distance); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_GIS__Distance__Fast_cosine_distance)
{
    dVAR; dXSARGS;
    if (items != 4)
       croak_xs_usage(cv,  "lat1, lon1, lat2, lon2");
    {
	double	lat1 = (double)SvNV(ST(0))
;
	double	lon1 = (double)SvNV(ST(1))
;
	double	lat2 = (double)SvNV(ST(2))
;
	double	lon2 = (double)SvNV(ST(3))
;
	double	RETVAL;
	dXSTARG;

	RETVAL = cosine_distance(lat1, lon1, lat2, lon2);
	TARGn((double)RETVAL, 1);
	ST(0) = TARG;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_GIS__Distance__Fast_great_circle_distance); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_GIS__Distance__Fast_great_circle_distance)
{
    dVAR; dXSARGS;
    if (items != 4)
       croak_xs_usage(cv,  "lat1, lon1, lat2, lon2");
    {
	double	lat1 = (double)SvNV(ST(0))
;
	double	lon1 = (double)SvNV(ST(1))
;
	double	lat2 = (double)SvNV(ST(2))
;
	double	lon2 = (double)SvNV(ST(3))
;
	double	RETVAL;
	dXSTARG;

	RETVAL = great_circle_distance(lat1, lon1, lat2, lon2);
	TARGn((double)RETVAL, 1);
	ST(0) = TARG;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_GIS__Distance__Fast_haversine_distance); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_GIS__Distance__Fast_haversine_distance)
{
    dVAR; dXSARGS;
    if (items != 4)
       croak_xs_usage(cv,  "lat1, lon1, lat2, lon2");
    {
	double	lat1 = (double)SvNV(ST(0))
;
	double	lon1 = (double)SvNV(ST(1))
;
	double	lat2 = (double)SvNV(ST(2))
;
	double	lon2 = (double)SvNV(ST(3))
;
	double	RETVAL;
	dXSTARG;

	RETVAL = haversine_distance(lat1, lon1, lat2, lon2);
	TARGn((double)RETVAL, 1);
	ST(0) = TARG;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_GIS__Distance__Fast_null_distance); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_GIS__Distance__Fast_null_distance)
{
    dVAR; dXSARGS;
    if (items != 4)
       croak_xs_usage(cv,  "lat1, lon1, lat2, lon2");
    {
	double	lat1 = (double)SvNV(ST(0))
;
	double	lon1 = (double)SvNV(ST(1))
;
	double	lat2 = (double)SvNV(ST(2))
;
	double	lon2 = (double)SvNV(ST(3))
;
	double	RETVAL;
	dXSTARG;

	RETVAL = null_distance(lat1, lon1, lat2, lon2);
	TARGn((double)RETVAL, 1);
	ST(0) = TARG;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_GIS__Distance__Fast_polar_distance); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_GIS__Distance__Fast_polar_distance)
{
    dVAR; dXSARGS;
    if (items != 4)
       croak_xs_usage(cv,  "lat1, lon1, lat2, lon2");
    {
	double	lat1 = (double)SvNV(ST(0))
;
	double	lon1 = (double)SvNV(ST(1))
;
	double	lat2 = (double)SvNV(ST(2))
;
	double	lon2 = (double)SvNV(ST(3))
;
	double	RETVAL;
	dXSTARG;

	RETVAL = polar_distance(lat1, lon1, lat2, lon2);
	TARGn((double)RETVAL, 1);
	ST(0) = TARG;
    }
    XSRETURN(1);
}


XS_EUPXS(XS_GIS__Distance__Fast_vincenty_distance); /* prototype to pass -Wmissing-prototypes */
XS_EUPXS(XS_GIS__Distance__Fast_vincenty_distance)
{
    dVAR; dXSARGS;
    if (items != 4)
       croak_xs_usage(cv,  "lat1, lon1, lat2, lon2");
    {
	double	lat1 = (double)SvNV(ST(0))
;
	double	lon1 = (double)SvNV(ST(1))
;
	double	lat2 = (double)SvNV(ST(2))
;
	double	lon2 = (double)SvNV(ST(3))
;
	double	RETVAL;
	dXSTARG;

	RETVAL = vincenty_distance(lat1, lon1, lat2, lon2);
	TARGn((double)RETVAL, 1);
	ST(0) = TARG;
    }
    XSRETURN(1);
}

#ifdef __cplusplus
extern "C" {
#endif
XS_EXTERNAL(boot_GIS__Distance__Fast); /* prototype to pass -Wmissing-prototypes */
XS_EXTERNAL(boot_GIS__Distance__Fast)
{
#if PERL_VERSION_LE(5, 21, 5)
    dVAR; dXSARGS;
#else
    dVAR; dXSBOOTARGSXSAPIVERCHK;
#endif
#if PERL_VERSION_LE(5, 8, 999) /* PERL_VERSION_LT is 5.33+ */
    char* file = __FILE__;
#else
    const char* file = __FILE__;
#endif

    PERL_UNUSED_VAR(file);

    PERL_UNUSED_VAR(cv); /* -W */
    PERL_UNUSED_VAR(items); /* -W */
#if PERL_VERSION_LE(5, 21, 5)
    XS_VERSION_BOOTCHECK;
#  ifdef XS_APIVERSION_BOOTCHECK
    XS_APIVERSION_BOOTCHECK;
#  endif
#endif

        newXS_deffile("GIS::Distance::Fast::alt_distance", XS_GIS__Distance__Fast_alt_distance);
        newXS_deffile("GIS::Distance::Fast::cosine_distance", XS_GIS__Distance__Fast_cosine_distance);
        newXS_deffile("GIS::Distance::Fast::great_circle_distance", XS_GIS__Distance__Fast_great_circle_distance);
        newXS_deffile("GIS::Distance::Fast::haversine_distance", XS_GIS__Distance__Fast_haversine_distance);
        newXS_deffile("GIS::Distance::Fast::null_distance", XS_GIS__Distance__Fast_null_distance);
        newXS_deffile("GIS::Distance::Fast::polar_distance", XS_GIS__Distance__Fast_polar_distance);
        newXS_deffile("GIS::Distance::Fast::vincenty_distance", XS_GIS__Distance__Fast_vincenty_distance);
#if PERL_VERSION_LE(5, 21, 5)
#  if PERL_VERSION_GE(5, 9, 0)
    if (PL_unitcheckav)
        call_list(PL_scopestack_ix, PL_unitcheckav);
#  endif
    XSRETURN_YES;
#else
    Perl_xs_boot_epilog(aTHX_ ax);
#endif
}

#ifdef __cplusplus
}
#endif
